/****************************************************************************
 *                                                                          *
 * The contents of this file are subject to the WebStone Public License     *
 * Version 1.0 (the "License"); you may not use this file except in         *
 * compliance with the License. You may obtain a copy of the License        *
 * at http://www.mindcraft.com/webstone/license10.html                      *
 *                                                                          *
 * Software distributed under the License is distributed on an "AS IS"      *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See      *
 * the License for the specific language governing rights and limitations   *
 * under the License.                                                       *
 *                                                                          *
 * The Original Code is WebStone 2.5.                                       *
 *                                                                          *
 * The Initial Developer of the Original Code is Silicon Graphics, Inc.     *
 * and Mindcraft, Inc.. Portions created by Silicon Graphics. and           *
 * Mindcraft. are Copyright (C) 1995-1998 Silicon Graphics, Inc. and        *
 * Mindcraft, Inc. All Rights Reserved.                                     *
 *                                                                          *
 * Contributor(s): ______________________________________.                  *
 *                                                                          *
 * @(#) bench.c 2.4@(#)                                                     *
 ***************************************************************************/


#include "sysdep.h"
#include "bench.h"

#ifdef UNIX_THREADS

#include <pthread.h>
#include <stdlib.h>
#include <errno.h>

pthread_once_t tls_once = PTHREAD_ONCE_INIT;
pthread_key_t tls_key;

static void	tls_key_init(void);

/* Initialize the per-thread key. */
static void
tls_key_init(void)
{

	if (pthread_key_create(&tls_key, NULL) != 0)
		exit(1);
}

/* Return per-thread instance data. */
struct thread_data *
tls_getcb(void)
{
	struct thread_data *tdp;

	if (pthread_once(&tls_once, tls_key_init) != 0)
		exit(1);

	tdp = pthread_getspecific(tls_key);
	if (tdp == NULL) {
		tdp = calloc(1, sizeof(*tdp));
		if (tdp == NULL)
			errexit("failed to alloc thread data, aborting\n");
		if (pthread_setspecific(tls_key, tdp) != 0)
			errexit("failed to set thread data, aborting\n");
		tdp->td_mastersock = BADSOCKET_VALUE;
		tdp->td_debugfile = stderr;
	}
	return (tdp);
}

int
InterlockedIncrement(int *val)
{
	static pthread_mutex_t interlock_mtx = PTHREAD_MUTEX_INITIALIZER;
	int rv;

	pthread_mutex_lock(&interlock_mtx);
	rv = *val;
	(*val)++;
	pthread_mutex_unlock(&interlock_mtx);
	return (rv);
}

int
_beginthread(void (*start_routine)(void *), int stack_size, void *arg)
{
	pthread_t dummy;
	pthread_attr_t attr;
	int err;

	err = pthread_attr_init(&attr);
	if (err)
		goto errout;
	err = pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	if (err)
		goto errout;
	if (stack_size != 0 &&
	    (err = pthread_attr_setstacksize(&attr, stack_size)) != 0)
		goto errout;

	err = pthread_create(&dummy, &attr,
	    (void *(*)(void *))start_routine, arg);
	if (err)
		goto errout;
	pthread_attr_destroy(&attr);
	return (0);
errout:
	errno = err;
	return (-1);
}

int
ReleaseSemaphore(HANDLE sem, long count, long *prev)
{
	int error = 0;

	pthread_mutex_lock(&sem->ms_mtx);
	if (prev != NULL)
		*prev = sem->ms_val;
	if (sem->ms_val + count > sem->ms_max) {
		errno = EINVAL;
		error = -1;
	} else {
		while (count--) {
			pthread_cond_signal(&sem->ms_cond);
			sem->ms_val++;
		}
	}
	pthread_mutex_unlock(&sem->ms_mtx);
	return (error);
}

HANDLE
CreateSemaphore(void *attr, long initialcount, long maxcount, const char *name)
{
	HANDLE ret;

	ret = malloc(sizeof(HANDLE));
	if (ret == NULL)
		return (NULL);
	if (pthread_mutex_init(&ret->ms_mtx, NULL) != 0)
		goto err;
	if (pthread_cond_init(&ret->ms_cond, NULL) != 0) {
		pthread_mutex_destroy(&ret->ms_mtx);
		goto err;
	}
	ret->ms_val = initialcount;
	ret->ms_max = maxcount;
	return (ret);
err:
	free(ret);
	return (NULL);
}

void
CloseHandle(HANDLE sem)
{

	pthread_mutex_destroy(&sem->ms_mtx);
	pthread_cond_destroy(&sem->ms_cond);
	free(sem);
}

int
WaitForSingleObject(HANDLE sem, int time)
{
	struct timespec ts;
	int error = 0;

	ts.tv_sec = time / 1000000L;
	ts.tv_nsec = time * 1000;

	pthread_mutex_lock(&sem->ms_mtx);
	while (error == 0 && sem->ms_val == 0) {
		if (time == INFINITE)
			pthread_cond_wait(&sem->ms_cond, &sem->ms_mtx);
		else
			error = pthread_cond_timedwait(&sem->ms_cond,
			    &sem->ms_mtx, &ts);
	}
	sem->ms_val--;
	pthread_mutex_unlock(&sem->ms_mtx);
	if (error == ETIMEDOUT)
		return (WAIT_TIMEOUT);
	if (error == 0)
		return (WAIT_OBJECT_0);
	errexit("WaitForSingleObject unexpected error\n", error);
	return (-1);
}

#endif

/* strerror() */
#ifndef HAVE_STRERROR
/* strerror is not available on SunOS 4.1.3 and others */
extern int sys_nerr;
extern char *sys_errlist[];
extern int errno;

char *strerror(int errnum)
{ 

    if (errnum<sys_nerr)
    {
	return(sys_errlist[errnum]);
    }

    return(NULL);
}

#endif /* strerror() */


/* stub routines for NT */ 

#ifdef WIN32
#include <winsock.h>
#include <process.h>

int getpid(void) {

    return GetCurrentThreadId();
}

void sleep(int sec) {

    Sleep(sec*1000);
}
#endif /* WIN32 */

